/*
  Copyright 2018-2021, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2021, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  This file is part of GekkoFS.

  GekkoFS is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  GekkoFS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GekkoFS.  If not, see <https://www.gnu.org/licenses/>.

  SPDX-License-Identifier: GPL-3.0-or-later
*/

#include <fstream>
#include <regex>
#include <random>
#include <utility>

#include <hermes.hpp>

#include <api/info.hpp>
#include <api/connection.hpp>
#include <api/detail/connection_impl.hpp>

#include <client/rpc/rpc_types.hpp> // TODO: rpc_types.hpp might need to move
#include <global/rpc/rpc_util.hpp>
#include <global/path_util.hpp>


#include <global/rpc/distributor.hpp>
#include <global/metadata.hpp>
#include <global/arithmetic/arithmetic.hpp>


////////////////////////////////////////////////////////////////////////////////
// helper types and functions                                                 //
////////////////////////////////////////////////////////////////////////////////
namespace {

using namespace gkfs::api;

// TODO: mostly identical to load_hostfile() in preload_util.cpp (could not
//  reuse due to CTX modification in extract_protocol(), which is called from
// load_hostfile()... Consider refactoring.
// std::tuple<rpc::config, std::vector<std::pair<std::string, std::string>>>

std::vector<fs::host>
load_hostfile(const std::string_view& hostfile) {

    //    LOG(DEBUG, "Loading hosts_ file: \"{}\"", hostfile);

    std::ifstream lf(std::string{hostfile});
    if(!lf) {
        throw std::runtime_error(fmt::format("{}", strerror(errno)));
    }
    // std::vector<std::pair<std::string, std::string>> hosts_;
    std::vector<fs::host> hosts;
    const std::regex line_re(R"(^(\S+)\s+(\S+)$)",
                             std::regex::ECMAScript | std::regex::optimize);
    std::string line;
    std::string host;
    std::string uri;
    std::smatch match;
    while(getline(lf, line)) {
        if(!regex_match(line, match, line_re)) {

            //            LOG(ERROR, "Unrecognized line format: [hostfile: '{}',
            //            line: '{}']",
            //                hostfile, line);

            throw std::runtime_error(fmt::format("invalid line: '{}'", line));
        }
        host = match[1];
        uri = match[2];
        hosts.emplace_back(host, uri);
    }
    if(hosts.empty()) {
        throw std::runtime_error(
                "hostfile exists but no addresses could be extracted");
    }

    return hosts;
}

std::vector<fs::host>
find_hosts(const std::string_view& hostfile) {
    try {
        return ::load_hostfile(hostfile);
    } catch(const std::exception& ex) {
        throw std::runtime_error(fmt::format("Error loading hostfile '{}': {}",
                                             hostfile, ex.what()));
    }
}

rpc::config
fetch_rpc_config(const std::string& uri) {

    rpc::config rpc_info;

    if(uri.rfind("://") == std::string::npos) {
        // invalid format. kill client
        throw std::runtime_error(
                fmt::format("Invalid format for URI: '{}'", uri));
    }
    std::string protocol{};

    if(uri.find(gkfs::rpc::protocol::ofi_sockets) != std::string::npos) {
        protocol = gkfs::rpc::protocol::ofi_sockets;
    } else if(uri.find(gkfs::rpc::protocol::ofi_psm2) != std::string::npos) {
        protocol = gkfs::rpc::protocol::ofi_psm2;
    } else if(uri.find(gkfs::rpc::protocol::ofi_verbs) != std::string::npos) {
        protocol = gkfs::rpc::protocol::ofi_verbs;
    }
    // check for shared memory protocol. Can be plain shared memory or real ofi
    // protocol + auto_sm
    if(uri.find(gkfs::rpc::protocol::na_sm) != std::string::npos) {
        if(protocol.empty())
            protocol = gkfs::rpc::protocol::na_sm;
        else
            rpc_info.auto_sm_enabled_ = true;
    }
    if(protocol.empty()) {
        // unsupported protocol. kill client
        throw std::runtime_error(fmt::format(
                "Unsupported RPC protocol found in hosts_ file with URI: '{}'",
                uri));
    }
    //    LOG(INFO,
    //        "RPC protocol '{}' extracted from hosts_ file. Using auto_sm is
    //        '{}'", protocol, CTX->auto_sm());
    rpc_info.protocol_string_ = protocol;

    return rpc_info;
}


/**
 * Looks up a host endpoint via Hermes
 * @param uri
 * @param max_retries
 * @return hermes endpoint, if successful
 * @throws std::runtime_error
 */
// TODO: almost identical to lookup_endpoint() in preload_util.cpp, generalize
hermes::endpoint
lookup_endpoint(const std::unique_ptr<hermes::async_engine>& rpc_service,
                const std::string_view& uri, std::size_t max_retries = 3) {

    //    LOG(DEBUG, "Looking up address \"{}\"", uri);

    std::random_device rd; // obtain a random number from hardware
    std::size_t attempts = 0;
    std::string error_msg;

    do {
        try {
            return rpc_service->lookup(std::string{uri});
        } catch(const std::exception& ex) {
            error_msg = ex.what();

            // LOG(WARNING, "Failed to lookup address '{}'. Attempts [{}/{}] ",
            //    uri, attempts + 1, max_retries);

            // Wait a random amount of time and try again
            std::mt19937 g(rd()); // seed the random generator
            std::uniform_int_distribution<> distr(
                    50, 50 * (attempts + 2)); // define the range
            std::this_thread::sleep_for(std::chrono::milliseconds(distr(g)));
            continue;
        }
    } while(++attempts < max_retries);

    throw std::runtime_error(
            fmt::format("Endpoint for address '{}' could not be found ({})",
                        uri, error_msg));
}


// TODO: mostly identical to init_hermes_client() in preload.cpp
// TODO: consider refactoring them
std::unique_ptr<hermes::async_engine>
create_rpc_service(bool use_auto_sm, const std::string_view& protocol_string) {

    try {

        hermes::engine_options opts{};

        if(use_auto_sm) {
            opts |= hermes::use_auto_sm;
        }

        if(gkfs::rpc::protocol::ofi_psm2 == protocol_string) {
            opts |= hermes::force_no_block_progress;
        }

        return std::make_unique<hermes::async_engine>(
                hermes::get_transport_type(std::string{protocol_string}), opts);
    } catch(const std::exception& ex) {
        throw std::runtime_error(fmt::format(
                "Failed to initialize RPC subsystem: {}\n", ex.what()));
    }
}

std::unique_ptr<hermes::async_engine>
start_rpc_service(const rpc::config& rpc_config) {
    auto rpc_service = create_rpc_service(rpc_config.auto_sm_enabled_,
                                          rpc_config.protocol_string_);
    rpc_service->run();
    return rpc_service;
}

// TODO: mostly identical to connec_to_hosts() in rpc_util.cpp
fs::detail::peer_list
resolve(const std::unique_ptr<hermes::async_engine>& rpc_service,
        const std::vector<fs::host>& hosts) {

    const auto localhost_name = gkfs::rpc::get_my_hostname(true);
    uint64_t localhost_id = 0;
    bool localhost_found = false;

    std::vector<hermes::endpoint> endpoints;
    endpoints.resize(hosts.size());

    std::vector<uint64_t> host_ids(hosts.size());
    // populate vector with [0, ..., host_size - 1]
    std::iota(std::begin(host_ids), std::end(host_ids), 0);
    /*
     * Shuffle hosts_ to balance addr lookups to all hosts_
     * Too many concurrent lookups send to same host
     * could overwhelm the server,
     * returning error when addr lookup
     */
    std::random_device rd; // obtain a random number from hardware
    std::mt19937 g(rd());  // seed the random generator
    std::shuffle(host_ids.begin(), host_ids.end(), g); // Shuffle hosts_
    // vector
    // lookup addresses and put abstract server addresses into rpc_addresses

    for(const auto& id : host_ids) {
        const auto& hostname = hosts.at(id).hostname();
        const auto& uri = hosts.at(id).uri();

        endpoints[id] = ::lookup_endpoint(rpc_service, uri);

        if(!localhost_found && hostname == localhost_name) {
            //                LOG(DEBUG, "Found local host: {}", hostname);
            localhost_id = id;
            localhost_found = true;
        }

        //            LOG(DEBUG, "Found peer: {}",
        //            endpoints[id].to_string());
    }

    return fs::detail::peer_list{localhost_id, endpoints};
}

// TODO: mostly identical to gkfs::rpc::forward_get_fs_config() in
// TODO: forward_management.cpp, consider refactoring them
fs::config
fetch_fs_config(const std::unique_ptr<hermes::async_engine>& rpc_service,
                const fs::detail::peer_list& peers) {

    const auto endp = peers.localhost();
    gkfs::rpc::fs_config::output out;

    try {
        //        LOG(DEBUG, "Retrieving file system configurations from
        //        daemon");
        // TODO(amiranda): add a post() with RPC_TIMEOUT to hermes so that
        // we can retry for RPC_TRIES (see old commits with margo)
        // TODO(amiranda): hermes will eventually provide a post(endpoint)
        // returning one result and a broadcast(endpoint_set) returning a
        // result_set. When that happens we can remove the .at(0) :/
        out = rpc_service->post<gkfs::rpc::fs_config>(endp).get().at(0);
    } catch(const std::exception& ex) {
        throw std::runtime_error(
                "Error retrieving gekkofs_instance configurations from "
                "daemon");
    }

    return {out.atime_state(),  out.mtime_state(),
            out.ctime_state(),  out.link_cnt_state(),
            out.blocks_state(), out.uid(),
            out.gid(),          out.rootdir(),
            out.mountdir()};
}


// TODO: mostly identical to forward_create in forward_metadata.cpp; refactor?


} // namespace


namespace gkfs::api::fs {

connection::impl::impl(const std::string_view& hostfile)
    : hosts_(find_hosts(hostfile)),
      rpc_config_(fetch_rpc_config(hosts_.at(0).uri())),
      rpc_service_(start_rpc_service(rpc_config_)),
      peers_(resolve(rpc_service_, hosts_)),
      fs_config_(fetch_fs_config(rpc_service_, peers_)),
      distributor_(std::make_unique<gkfs::rpc::SimpleHashDistributor>(
              peers_.localhost_id(), peers_.size())) {

    using namespace std::string_literals;

#if 0
    // read host information
        std::vector<std::pair<std::string, std::string>> hosts_{};

        try {
            //        LOG(INFO, "Loading peer addresses...");
            std::tie(rpc_config_, hosts_) = ::find_hosts();
        } catch(const std::exception& e) {
            throw std::runtime_error(
                    fmt::format("Failed to load hosts_ addresses:"
                                " {}",
                                e.what()));
        }
#endif

    // rpc_service_ = create_rpc_service();
    // rpc_service_->run();

#if 0
    try {
            ping(hosts_);
        } catch(const std::exception& e) {
            throw std::runtime_error(
                    fmt::format("Failed to connect to hosts_: {}\n", e.what()));
        }
#endif

#if 0
    distributor_ = std::make_unique<gkfs::rpc::SimpleHashDistributor>(
                local_host_id_, host_endpoints_.size());
#endif

    // LOG(INFO, "Retrieving file system configuration...");
#if 0
    try {
            fs_config_ = fetch_fs_info();
        } catch(const std::exception& ex) {
            throw std::runtime_error(
                    fmt::format("Unable to fetch file system "
                                "configurations from daemon process "
                                "through RPC: {}",
                                ex.what()));
        }
#endif

    // LOG(INFO, "Environment initialization successful.");
}

// connection::impl::impl(const impl& other) = delete;
connection::impl::impl(impl&& rhs) noexcept = default;
// connection::impl&
// connection::impl::operator=(const impl& other) = delete;
connection::impl&
connection::impl::operator=(impl&& rhs) noexcept = default;
connection::impl::~impl() = default;

std::size_t
connection::impl::put(const void* buffer, size_t size,
                      const std::filesystem::path& gkfs_filename,
                      off_t start_offset) const {

    create(gkfs_filename, 0660 | S_IFREG);
    return pwrite(gkfs_filename, buffer, size, start_offset);
}

/**
 * Send an RPC for a stat request
 *
 * @param path
 * @param attr
 * @return the attributes requested
 */
std::string
connection::impl::stat(const std::string& filename) const {

    auto endp = peers_.at(distributor_->locate_file_metadata(filename));

    //        LOG(DEBUG, "Sending RPC ...");
    // TODO(amiranda): add a post() with RPC_TIMEOUT to hermes so that we
    // can retry for RPC_TRIES (see old commits with margo)
    // TODO(amiranda): hermes will eventually provide a post(endpoint)
    // returning one result and a broadcast(endpoint_set) returning a
    // result_set. When that happens we can remove the .at(0) :/
    auto out = rpc_service_->post<gkfs::rpc::stat>(endp, filename).get().at(0);
    //        LOG(DEBUG, "Got response success: {}", out.err());

    if(int error_code = out.err()) {
        throw std::runtime_error(fmt::format("Failed to get metadata for "
                                             "'{}': {}",
                                             filename, strerror(error_code)));
    }

    return out.db_val();
}


// TODO: mostly identical to gkfs_create in gkfs_functions.cpp;
//  refactor?
void
connection::impl::create(const std::string& gkfs_filename, mode_t mode) const {

    // file type must be set
    switch(mode & S_IFMT) {
        case 0:
            mode |= S_IFREG;
            break;
        case S_IFREG: // intentionally fall-through
        case S_IFDIR:
            break;
        case S_IFCHR: // intentionally fall-through
        case S_IFBLK:
        case S_IFIFO:
        case S_IFSOCK:
            //            LOG(WARNING, "Unsupported node type");
            throw std::runtime_error(strerror(ENOTSUP));
        default:
            //            LOG(WARNING, "Unrecognized node type");
            throw std::runtime_error(strerror(EINVAL));
    }

    // TODO: the connection should allow the API to check if
    //       the daemons support CREATE_CHECK_PARENTS.
    //       Another option is to offer an option to disable the check
    // For now we are assuming it doesn't matter
    if(false) { // TODO: if(connection.options().fs_has_create_check_parents)...
        check_parent_directory(gkfs_filename);
    }

    if(auto err = forward_create_rpc(gkfs_filename, mode)) {
        throw std::runtime_error(
                fmt::format("Error creating file: {}", strerror(err)));
    }
}

// TODO: mostly identical to forward_write in forward_data.cpp, refactor
ssize_t
connection::impl::pwrite(const std::string& gkfs_filename, const void* buffer,
                         size_t size, off_t offset) const {

    if(size == 0) {
        return 0;
    }

    const auto [update_error, updated_size] =
            forward_update_metadentry_size_rpc(gkfs_filename, size, offset,
                                               false);

    if(update_error) {
        //        LOG(ERROR, "update_metadentry_size() failed with update_error
        //        '{}'", update_error);
        throw std::runtime_error(fmt::format("Error writing to file: {}",
                                             strerror(update_error)));
    }

    const auto [write_error, written_size] = forward_write_rpc(
            gkfs_filename, buffer, offset, size, updated_size, false);

    if(write_error) {
        //        LOG(WARNING, "gkfs::rpc::forward_write() failed with
        //        update_error
        //        '{}'", update_error);
        throw std::runtime_error(fmt::format("Error writing to file: {}",
                                             strerror(write_error)));
    }

    return written_size;
}

////////////////////////////////////////////////////////////////////////////////
// private helpers                                                            //
////////////////////////////////////////////////////////////////////////////////
/**
 * Checks if metadata for parent directory exists (can be disabled with
 * CREATE_CHECK_PARENTS). errno may be set
 * @param path
 * @return 0 on success, -1 on failure
 */
void
connection::impl::check_parent_directory(const std::string& path) const {

    auto p_comp = gkfs::path::dirname(path);
    auto md = fetch_metadata(p_comp, false);
    if(!md) {
        if(errno == ENOENT) {
            //            LOG(DEBUG, "Parent component does not exist: '{}'",
            //            p_comp);
            throw std::runtime_error(
                    fmt::format("Parent '{}' does not exist.", p_comp));
        } else {
            //            LOG(ERROR, "Failed to get metadata for parent
            //            component '{}': {}",
            //                path, strerror(errno));
            throw std::runtime_error(
                    fmt::format("Failed to get metadata for '{}': {}", p_comp,
                                strerror(errno)));
        }
    }

    if(!S_ISDIR(md->mode())) {
        //        LOG(DEBUG, "Parent component is not a directory: '{}'",
        //        p_comp);
        throw std::runtime_error(fmt::format("Parent '{}' is not a "
                                             "directory",
                                             p_comp));
    }
}

// TODO: very similar to get_metadata() from preload_util.hpp. Consider
//  refactoring
/**
 * Retrieve metadata from daemon and return Metadata object
 * errno may be set
 * @param path
 * @param follow_links
 * @return Metadata
 */
std::optional<gkfs::metadata::Metadata>
connection::impl::fetch_metadata(const std::string& path,
                                 bool follow_links) const {
    std::string attr;

    attr = stat(path);

    //#ifdef HAS_SYMLINKS
    if(false) { // TODO: if(connection.options().fs_has_symlinks)...
        if(follow_links) {
            gkfs::metadata::Metadata md{attr};
            while(md.is_link()) {
                attr = stat(md.target_path());
                return gkfs::metadata::Metadata{attr};
            }
        }
    }
    //#endif
    return gkfs::metadata::Metadata{attr};
}

// TODO: mostly identical to forward_create in forward_metadata.cpp; refactor?
int
connection::impl::forward_create_rpc(const std::string& gkfs_filename,
                                     const mode_t mode) const {

    const auto endp =
            peers_.at(distributor_->locate_file_metadata(gkfs_filename));

    try {
        //        LOG(DEBUG, "Sending RPC ...");
        // TODO(amiranda): add a post() with RPC_TIMEOUT to hermes so that we
        // can retry for RPC_TRIES (see old commits with margo)
        // TODO(amiranda): hermes will eventually provide a post(endpoint)
        // returning one result and a broadcast(endpoint_set) returning a
        // result_set. When that happens we can remove the .at(0) :/
        auto out =
                rpc_service_->post<gkfs::rpc::create>(endp, gkfs_filename, mode)
                        .get()
                        .at(0);
        //        LOG(DEBUG, "Got response success: {}", out.err());

        return out.err();
    } catch(const std::exception& ex) {
        //        LOG(ERROR, "while getting rpc output");
        throw std::runtime_error(fmt::format("Error retrieving RPC output: "
                                             "{}",
                                             ex.what()));
    }
}

// TODO: mostly identical to forward_update_metadentry_size_rpc, refactor
std::tuple<int, off64_t>
connection::impl::forward_update_metadentry_size_rpc(
        const std::string& gkfs_filename, size_t size, off_t offset,
        bool is_append) const {

    const auto endp =
            peers_.at(distributor_->locate_file_metadata(gkfs_filename));

    try {
        //        LOG(DEBUG, "Sending RPC ...");
        // TODO(amiranda): add a post() with RPC_TIMEOUT to hermes so that we
        // can retry for RPC_TRIES (see old commits with margo)
        // TODO(amiranda): hermes will eventually provide a post(endpoint)
        // returning one result and a broadcast(endpoint_set) returning a
        // result_set. When that happens we can remove the .at(0) :/
        auto out = rpc_service_
                           ->post<gkfs::rpc::update_metadentry_size>(
                                   endp, gkfs_filename, size, offset, is_append)
                           .get()
                           .at(0);

        //        LOG(DEBUG, "Got response success: {}", out.err());

        if(out.err())
            return {out.err(), 0};
        else
            return {0, out.ret_size()};
    } catch(const std::exception& ex) {
        throw std::runtime_error(fmt::format("Error retrieving RPC output: "
                                             "{}",
                                             ex.what()));
    }
}

// TODO: mostly identical to forward_write in forward_data.cpp, refactor
std::tuple<int, ssize_t>
connection::impl::forward_write_rpc(const std::string& gkfs_filename,
                                    const void* buffer, off_t start_offset,
                                    size_t write_size,
                                    size_t updated_metadentry_size,
                                    bool is_append) const {

    using namespace gkfs::utils::arithmetic;

    assert(write_size > 0);

    // Calculate chunkid boundaries and numbers so that daemons know in
    // which interval to look for chunks
    off64_t offset = is_append ? start_offset
                               : static_cast<off64_t>(updated_metadentry_size -
                                                      write_size);

    auto chnk_start = block_index(offset, gkfs::config::rpc::chunksize);
    auto chnk_end = block_index((offset + write_size) - 1,
                                gkfs::config::rpc::chunksize);

    // Collect all chunk ids within count that have the same destination so
    // that those are send in one rpc bulk transfer
    std::map<uint64_t, std::vector<uint64_t>> target_chnks{};
    // contains the target ids, used to access the target_chnks map.
    // First idx is chunk with potential offset
    std::vector<uint64_t> targets{};

    // targets for the first and last chunk as they need special treatment
    uint64_t chnk_start_target = 0;
    uint64_t chnk_end_target = 0;

    for(uint64_t chnk_id = chnk_start; chnk_id <= chnk_end; chnk_id++) {
        auto target = distributor_->locate_data(gkfs_filename, chnk_id);

        if(target_chnks.count(target) == 0) {
            target_chnks.insert(
                    std::make_pair(target, std::vector<uint64_t>{chnk_id}));
            targets.push_back(target);
        } else {
            target_chnks[target].push_back(chnk_id);
        }

        // set first and last chnk targets
        if(chnk_id == chnk_start) {
            chnk_start_target = target;
        }

        if(chnk_id == chnk_end) {
            chnk_end_target = target;
        }
    }

    // some helper variables for async RPC
    std::vector<hermes::mutable_buffer> bufseq{
            hermes::mutable_buffer{const_cast<void*>(buffer), write_size},
    };

    // expose user buffers so that they can serve as RDMA data sources
    // (these are automatically "unexposed" when the destructor is called)
    hermes::exposed_memory local_buffers;

    try {
        local_buffers =
                rpc_service_->expose(bufseq, hermes::access_mode::read_only);

    } catch(const std::exception& ex) {
        //        LOG(ERROR, "Failed to expose buffers for RMA");
        throw std::runtime_error(fmt::format("Error exposing buffers for "
                                             "RDMA: {}",
                                             ex.what()));
    }

    std::vector<hermes::rpc_handle<gkfs::rpc::write_data>> handles;

    // Issue non-blocking RPC requests and wait for the result later
    //
    // TODO(amiranda): This could be simplified by adding a vector of inputs
    // to async_engine::broadcast(). This would allow us to avoid manually
    // looping over handles as we do below
    for(const auto& target : targets) {

        // total chunk_size for target
        auto total_chunk_size =
                target_chnks[target].size() * gkfs::config::rpc::chunksize;

        // receiver of first chunk must subtract the offset from first chunk
        if(target == chnk_start_target) {
            total_chunk_size -=
                    block_overrun(offset, gkfs::config::rpc::chunksize);
        }

        // receiver of last chunk must subtract
        if(target == chnk_end_target &&
           !is_aligned(offset + write_size, gkfs::config::rpc::chunksize)) {
            total_chunk_size -= block_underrun(offset + write_size,
                                               gkfs::config::rpc::chunksize);
        }

        auto endp = peers_.at(target);

        try {

            //            LOG(DEBUG, "Sending RPC ...");

            gkfs::rpc::write_data::input in(
                    gkfs_filename,
                    // first offset in targets is the chunk with
                    // a potential offset
                    block_overrun(offset, gkfs::config::rpc::chunksize), target,
                    peers_.size(),
                    // number of chunks handled by that destination
                    target_chnks[target].size(),
                    // chunk start id of this write
                    chnk_start,
                    // chunk end id of this write
                    chnk_end,
                    // total size to write
                    total_chunk_size, local_buffers);

            // TODO(amiranda): add a post() with RPC_TIMEOUT to hermes so that
            // we can retry for RPC_TRIES (see old commits with margo)
            // TODO(amiranda): hermes will eventually provide a post(endpoint)
            // returning one result and a broadcast(endpoint_set) returning a
            // result_set. When that happens we can remove the .at(0) :/
            handles.emplace_back(
                    rpc_service_->post<gkfs::rpc::write_data>(endp, in));

            //            LOG(DEBUG,
            //                "host: {}, path: \"{}\", chunks: {}, size: {},
            //                offset: {}", target, path, in.chunk_n(),
            //                total_chunk_size, in.offset());

        } catch(const std::exception& ex) {
            //            LOG(ERROR,
            //                "Unable to send non-blocking rpc for "
            //                "path \"{}\" [peer: {}]",
            //                path, target);
            throw std::runtime_error(fmt::format("Error sending non-block RPC"
                                                 " for '{}' [peer: {}]: {}",
                                                 gkfs_filename, target,
                                                 ex.what()));
        }
    }

    // Wait for RPC responses and then get response and add it to out_size
    // which is the written size All potential outputs are served to free
    // resources regardless of errors, although an errorcode is set.
    auto err = 0;
    ssize_t out_size = 0;
    std::size_t idx = 0;

    for(const auto& h : handles) {
        try {
            // XXX We might need a timeout here to not wait forever for an
            // output that never comes?
            auto out = h.get().at(0);

            if(out.err() != 0) {
                //                LOG(ERROR, "Daemon reported error: {}",
                //                out.err());
                err = out.err();
            }

            out_size += static_cast<ssize_t>(out.io_size());

        } catch(const std::exception& ex) {
            //            LOG(ERROR, "Failed to get rpc output for path \"{}\"
            //            [peer: {}]",
            //                path, targets[idx]);
            throw std::runtime_error(fmt::format("Error fetching RPC "
                                                 "response for path '{}' "
                                                 "[peer: {}]: {}",
                                                 gkfs_filename, targets[idx],
                                                 ex.what()));
        }

        idx++;
    }

    /*
     * Typically file systems return the size even if only a part of it was
     * written. In our case, we do not keep track which daemon fully wrote its
     * workload. Thus, we always return size 0 on error.
     */
    return {err, err ? 0 : out_size};
}


} // namespace gkfs::api::fs