/*
  Copyright 2018-2021, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2021, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  This file is part of GekkoFS.

  GekkoFS is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  GekkoFS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GekkoFS.  If not, see <https://www.gnu.org/licenses/>.

  SPDX-License-Identifier: GPL-3.0-or-later
*/

#include <filesystem>
#include <fstream>
#include <regex>

#include <tests/config.hpp>
#include <catch2/catch.hpp>
#include <fmt/format.h>
#include <boost/process.hpp>
#include <utility>
#include <boost/uuid/uuid.hpp>            // uuid class
#include <boost/uuid/uuid_generators.hpp> // generators
#include <boost/uuid/uuid_io.hpp>         // streaming operators etc.

#include <api/gekkofs.hpp>
#include <config.hpp>

using namespace std::literals;
namespace bp = boost::process;
namespace boost_fs = boost::filesystem;
namespace std_fs = std::filesystem;

constexpr auto default_hostfile = "gkfs_hosts.txt";


// A temporary file with RAII removal
struct temporary_file {

    explicit temporary_file(std_fs::path filename)
        : filename_(std::move(filename)), ofs_(filename_) {}

    temporary_file(std_fs::path filename, const std::string_view& text)
        : filename_(std::move(filename)), ofs_(filename_) {
        write(text);
    }

    ~temporary_file() {
        ofs_.close();
        std_fs::remove(filename_);
    }

    void
    write(const std::string_view& text) {
        ofs_ << text;
        ofs_.flush();
    }

    void
    write(const std::vector<char>& data) {
        for(const auto n : data) {
            ofs_ << n;
        }
        ofs_.flush();
    }

    std_fs::path
    filename() const {
        return filename_;
    }

    std::size_t
    size() const {
        return std_fs::file_size(filename_);
    };

    std_fs::path filename_;
    std::ofstream ofs_;
};

struct gkfs_config {
    gkfs_config(std_fs::path cwd, std_fs::path hosts_file, std_fs::path rootdir,
                std_fs::path mountdir, std::string interface)
        : cwd_(std::move(cwd)), hosts_file_(std::move(hosts_file)),
          rootdir_(std::move(rootdir)), mountdir_(std::move(mountdir)),
          interface_(std::move(interface)) {}

    const std_fs::path cwd_;
    const std_fs::path hosts_file_;
    const std_fs::path rootdir_;
    const std_fs::path mountdir_;
    const std::string interface_;
};

// A temporary daemon with RAII shutdown and environment removal
struct test_daemon {

    static std::tuple<std_fs::path, long>
    generate_environment(bool unique_port = true) {

        static boost::uuids::random_generator generator;
        boost::uuids::uuid uuid = generator();
        const auto basedir = "test-" + to_string(uuid);

        // return 0 if a unique_port port was requested: this will force the kernel
        // to provide a port that is unused when the daemon tries to bind
        // to the interface. This helps avoid race conditions where two
        // daemons are assigned the same port when creating
        // several test_daemons concurrently
        if(unique_port) {
            return {basedir, 0};
        }

        // otherwise just return a randomly generated port
        boost::random::mt19937 rng;
        boost::random::uniform_int_distribution<> dist(1025, 65535);
        const long port = dist(rng);

        return {basedir, port};
    }

    explicit test_daemon(const gkfs_config& cfg)
        : cfg_(cfg),
          abs_rootdir(std_fs::absolute(std_fs::current_path() / cfg.rootdir_)),
          abs_mountdir(
                  std_fs::absolute(std_fs::current_path() / cfg.mountdir_)) {

        auto env = boost::this_process::environment();
        bp::environment env_ = env;
        env_["GKFS_LOG_LEVEL"] = "100";
        env_["GKFS_DAEMON_LOG_PATH"] = cfg.cwd_ / "test_daemon.log";
        env_["GKFS_HOSTS_FILE"] = cfg.hosts_file_;

        if(!std_fs::exists(abs_rootdir)) {
            std_fs::create_directories(abs_rootdir);
        }

        if(!std_fs::exists(abs_mountdir)) {
            std_fs::create_directories(abs_mountdir);
        }

        auto search_paths = boost::this_process::path();
        search_paths.emplace_back(boost_fs::path{cmake::binary_dir} /
                                  boost_fs::path{"src/daemon"});

        daemon_ =
                bp::child(bp::search_path("gkfs_daemon", search_paths),
                          bp::args({"--rootdir"s, abs_rootdir, "--mountdir"s,
                                    abs_mountdir, "--listen"s, cfg.interface_}),
                          env_);

        // since we don't have a way to check whether the daemon has finished
        // its startup process, wait for a 100ms, otherwise any connections
        // to it will fail
        std::this_thread::sleep_for(2s);
    }

    ~test_daemon() {
        ::kill(daemon_.native_handle(), SIGTERM);

        if(!daemon_.wait_for(5s)) {
            daemon_.terminate();
        }

        std_fs::remove_all(abs_rootdir);
        std_fs::remove_all(abs_mountdir);
    }

    int
    port() const {
        // parse the generated hostfile to see what port the kernel gave us
        std::ifstream ifs{cfg_.hosts_file_};

        if(!ifs) {
            throw std::runtime_error(fmt::format("Error opening hostfile "
                                                 "'{}': {}",
                                                 cfg_.hosts_file_.string(),
                                                 strerror(errno)));
        }

        const std::regex re(R"(^(\S+)\s+(\S+):(\d+)$)",
                            std::regex::ECMAScript | std::regex::optimize);

        std::string line;
        std::smatch match;

        while(getline(ifs, line)) {
            if(!regex_match(line, match, re)) {
                throw std::runtime_error(
                        fmt::format("unrecognized line format: '{}'", line));
            }

            return std::stoi(match[3].str());
        }

        throw std::runtime_error("Could not extract port from hostfile");
    }

    bp::child daemon_;
    gkfs_config cfg_;
    std_fs::path abs_rootdir;
    std_fs::path abs_mountdir;
};

struct md5sum {

    explicit md5sum(const std::filesystem::path& filename,
                    const std::optional<gkfs_config>& gkfs_config = {}) {

        auto env = boost::this_process::environment();
        bp::environment env_ = env;

        if(gkfs_config) {
            env_["LD_PRELOAD"] = std_fs::path{cmake::binary_dir} /
                                 "src/client/libgkfs_intercept.so";
            env_["LIBGKFS_HOSTS_FILE"] = gkfs_config->cwd_ / default_hostfile;
            env_["LIBGKFS_LOG_OUTPUT"] = gkfs_config->cwd_ / "test_client.log";
        }

        bp::ipstream is;
        proc_ = bp::child(bp::search_path("md5sum"), filename.string(),
                          bp::std_err > stdout, bp::std_out > is, env_);

        std::string line;

        while(proc_.running() && std::getline(is, line) && !line.empty()) {
            output_.push_back(line);
        }

        proc_.wait();
    }

    std::string
    operator()() const {

        if(proc_.exit_code()) {
            throw std::runtime_error(
                    fmt::format("ERROR: md5sum: {}", fmt::join(output_, "\n")));
        }

        if(output_.size() != 1) {
            throw std::runtime_error(fmt::format(
                    "ERROR: md5sum: program called with more than one file : {}",
                    fmt::join(output_, "\n")));
        }

        const std::regex regex(R"(^(\S+)\s+(\S+)$)",
                               std::regex::ECMAScript | std::regex::optimize);
        std::string host;
        std::string uri;
        std::smatch match;
        if(!regex_match(output_[0], match, regex)) {
            throw std::runtime_error(fmt::format("ERROR: md5sum: unknown "
                                                 "output format: {}",
                                                 output_[0]));
        }

        return match[1];
    }

    bp::child proc_;
    std::vector<std::string> output_;
};


std::vector<char>
generate_random_data(int n) {

    const std::string chars{
            "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"};

    boost::random::mt19937 rng;
    boost::random::uniform_int_distribution<> dist(
            0, static_cast<int>(chars.size() - 1));
    std::vector<char> data;
    data.reserve(n);

    for(int i = 0; i < n; ++i) {
        data.push_back(chars[dist(rng)]);
    }

    return data;
}

SCENARIO(" API connections to gkfs work as expected ", "[api][connection]") {

    // TODO: tests for connecting non-existing daemons are missing. The
    // reason
    // TODO: is that since we don't have timeouts yet in Hermes, if the
    // TODO: daemon is missing the connection RPC hangs forever while
    // waiting
    // TODO: for a reply that will never arrive

    using namespace std::string_literals;
    using Catch::Matchers::Contains;
    using namespace gkfs::api;

    GIVEN(" a fs::connection ") {

        WHEN(" hostfile does not exist ") {

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(fs::connection(),
                                    Contains("No such file or directory"s));
            }
        }

        WHEN(" hostfile is empty ") {

            temporary_file tmp{"empty.hostfile"};

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(
                        fs::connection(tmp.filename()),
                        Contains("hostfile exists but no addresses could be "
                                 "extracted"s));
            }
        }

        WHEN(" hostfile is invalid ") {

            temporary_file tmp{"invalid.hostfile",
                               "foobarbaz\nbarfoobaz\nbazbarfoo\n"};

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(fs::connection(tmp.filename()),
                                    Contains("invalid line: 'foobarbaz'"s));
            }
        }

        WHEN(" hostfile is partially invalid ") {

            temporary_file tmp{"pinvalid.hostfile",
                               "host.example.com "
                               "ofi+sockets://127.0.0.1:11586\n"
                               "foobarbaz\nbarfoobaz\nbazbarfoo\n"};

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(fs::connection(tmp.filename()),
                                    Contains("invalid line: 'foobarbaz'"s));
            }
        }

        WHEN(" hostfile is valid ") {

            const auto [testdir, port] = test_daemon::generate_environment();
            gkfs_config cfg{testdir, testdir / default_hostfile,
                            testdir / "rootdir", testdir / "mountdir",
                            fmt::format("lo:{}", port)};

            test_daemon daemon(cfg);

            temporary_file tmp{
                    "valid.hostfile",
                    fmt::format("localhost ofi+sockets://127.0.0.1:{}\n",
                                daemon.port())};

            THEN(" a connection instance is constructed ") {
                REQUIRE_NOTHROW(fs::connection(testdir / default_hostfile));
            }
        }
    }
}

SCENARIO(" Staging files with a standalone connection as expected ",
         "[api][staging][put][standalone]") {

    using namespace gkfs::api;

    GIVEN(" an active gkfs daemon ") {

        const auto [testdir, port] = test_daemon::generate_environment();

        gkfs_config cfg{testdir, testdir / default_hostfile,
                        testdir / "rootdir", testdir / "mountdir",
                        fmt::format("lo:{}", port)};
        test_daemon daemon(cfg);

        AND_GIVEN(" no previously established connection to the daemon ") {
            WHEN(" staging a file into gkfs ") {
                AND_WHEN(" the file size equals 0 ") {

                    int file_size = 0;

                    CAPTURE(file_size);

                    // generate a file with random data
                    temporary_file f{testdir / "vfs_testfile"};
                    std::vector<char> data = generate_random_data(file_size);
                    f.write(data);

                    // make sure the API connects to our test_daemon
                    auto env = boost::this_process::environment();
                    env["GKFS_API_HOSTS_FILE"] = testdir / default_hostfile;

                    // upload the file
                    auto bytes_written =
                            staging::put(static_cast<const void*>(data.data()),
                                         data.size(), "/gkfs_testfile", 0);

                    THEN(" no data is written ") {
                        REQUIRE(bytes_written == 0);
                    }
                }

                AND_WHEN(" the file size is not 0 ") {
                    auto cs = gkfs::config::rpc::chunksize;

                    int file_size =
                            GENERATE_COPY(5, 100, cs, std::floor(cs * 1.5),
                                          cs * 2 - 1, cs * 2, 500000);
                    CAPTURE(file_size);

                    // generate a file with random data
                    temporary_file f{testdir / "vfs_testfile"};
                    std::vector<char> data = generate_random_data(file_size);
                    f.write(data);

                    // make sure the API connects to our test_daemon
                    auto env = boost::this_process::environment();
                    env["GKFS_API_HOSTS_FILE"] = testdir / default_hostfile;

                    // upload the file
                    auto bytes_written =
                            staging::put(static_cast<const void*>(data.data()),
                                         data.size(), "/gkfs_testfile", 0);

                    THEN(" data contents are identical ") {
                        const auto vfs_digest = md5sum(f.filename())();
                        const auto gkfs_digest =
                                md5sum(cfg.mountdir_ / "gkfs_testfile", cfg)();
                        REQUIRE(vfs_digest == gkfs_digest);
                    }

                    AND_THEN(
                            " the number of written bytes reported is correct") {
                        REQUIRE(bytes_written == data.size());
                    }
                }
            }
        }
    }
}

SCENARIO(" Staging files reusing a connection works as expected ",
         "[api][staging][put][reusable]") {

    using namespace gkfs::api;

    GIVEN(" an active gkfs daemon ") {

        const auto [testdir, port] = test_daemon::generate_environment();

        gkfs_config cfg{testdir, testdir / default_hostfile,
                        testdir / "rootdir", testdir / "mountdir",
                        fmt::format("lo:{}", port)};
        test_daemon daemon(cfg);

        AND_GIVEN(" an already established connection to the daemon ") {

            temporary_file tmp{
                    "hostfile",
                    fmt::format("localhost ofi+sockets://127.0.0.1:{}\n",
                                daemon.port())};

            fs::connection fsc{tmp.filename()};

            WHEN(" staging a file into gkfs ") {

                AND_WHEN(" the file size equals 0 ") {

                    int file_size = 0;

                    CAPTURE(file_size);

                    // generate a file with random data
                    temporary_file f{testdir / "vfs_testfile"};
                    std::vector<char> data = generate_random_data(file_size);
                    f.write(data);

                    // make sure the API connects to our test_daemon
                    auto env = boost::this_process::environment();
                    env["GKFS_API_HOSTS_FILE"] = testdir / default_hostfile;

                    // upload the file
                    auto bytes_written = staging::put(
                            fsc, static_cast<const void*>(data.data()),
                            data.size(), "/gkfs_testfile", 0);

                    THEN(" no data is written ") {
                        REQUIRE(bytes_written == 0);
                    }
                }

                AND_WHEN(" the file size is not 0 ") {
                    auto cs = gkfs::config::rpc::chunksize;

                    int file_size =
                            GENERATE_COPY(5, 100, cs, std::floor(cs * 1.5),
                                          cs * 2 - 1, cs * 2, 500000);
                    CAPTURE(file_size);

                    // generate a file with random data
                    temporary_file f{testdir / "vfs_testfile"};
                    std::vector<char> data = generate_random_data(file_size);
                    f.write(data);

                    // make sure the API connects to our test_daemon
                    auto env = boost::this_process::environment();
                    env["GKFS_API_HOSTS_FILE"] = testdir / default_hostfile;

                    // upload the file
                    auto bytes_written = staging::put(
                            fsc, static_cast<const void*>(data.data()),
                            data.size(), "/gkfs_testfile", 0);

                    THEN(" data contents are identical ") {
                        const auto vfs_digest = md5sum(f.filename())();
                        const auto gkfs_digest =
                                md5sum(cfg.mountdir_ / "gkfs_testfile", cfg)();
                        REQUIRE(vfs_digest == gkfs_digest);
                    }

                    AND_THEN(
                            " the number of written bytes reported is correct") {
                        REQUIRE(bytes_written == data.size());
                    }
                }
            }
        }
    }
}